@@ -1,34 +1,37 @@
 /****************************************************************************
 **
-** Copyright (C) 2015 The Qt Company Ltd.
-** Contact: http://www.qt.io/licensing/
+** Copyright (C) 2017 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtWaylandCompositor module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL3$
+** $QT_BEGIN_LICENSE:LGPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
 ** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see http://www.qt.io/terms-conditions. For further
-** information use the contact form at http://www.qt.io/contact-us.
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
 ** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPLv3 included in the
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
 ** packaging of this file. Please review the following information to
 ** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl.html.
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or later as published by the Free
-** Software Foundation and appearing in the file LICENSE.GPL included in
-** the packaging of this file. Please review the following information to
-** ensure the GNU General Public License version 2.0 requirements will be
-** met: http://www.gnu.org/licenses/gpl-2.0.html.
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -39,14 +42,20 @@
 #include "qwaylandquicksurface.h"
 #include "qwaylandinputmethodcontrol.h"
 #include "qwaylandtextinput.h"
+#include "qwaylandquickoutput.h"
 #include <QtWaylandCompositor/qwaylandcompositor.h>
-#include <QtWaylandCompositor/qwaylandinput.h>
+#include <QtWaylandCompositor/qwaylandseat.h>
 #include <QtWaylandCompositor/qwaylandbufferref.h>
+#if QT_CONFIG(draganddrop)
+#include <QtWaylandCompositor/QWaylandDrag>
+#endif
 #include <QtWaylandCompositor/private/qwlclientbufferintegration_p.h>
 
 #include <QtGui/QKeyEvent>
 #include <QtGui/QGuiApplication>
 #include <QtGui/QScreen>
+#include <QtGui/QOpenGLFunctions>
+#include <QtGui/QOpenGLTexture>
 
 #include <QtQuick/QSGSimpleTextureNode>
 #include <QtQuick/QQuickWindow>
@@ -185,52 +194,40 @@ QWaylandBufferMaterial::QWaylandBufferMaterial(QWaylandBufferRef::BufferFormatEg
 
 QWaylandBufferMaterial::~QWaylandBufferMaterial()
 {
-    QOpenGLFunctions *gl = QOpenGLContext::currentContext()->functions();
-
-    for (GLuint texture : m_textures)
-        gl->glDeleteTextures(1, &texture);
 }
 
-void QWaylandBufferMaterial::setTextureForPlane(int plane, uint texture)
+void QWaylandBufferMaterial::setTextureForPlane(int plane, QOpenGLTexture *texture)
 {
     if (plane < 0 || plane >= bufferTypes[m_format].planeCount) {
         qWarning("plane index is out of range");
         return;
     }
 
-    QOpenGLFunctions *gl = QOpenGLContext::currentContext()->functions();
-    const GLenum target = bufferTypes[m_format].textureTarget;
-
-    gl->glBindTexture(target, texture);
-    setTextureParameters(target);
+    texture->bind();
+    setTextureParameters(texture->target());
 
     ensureTextures(plane - 1);
 
-    if (m_textures.size() <= plane) {
+    if (m_textures.size() <= plane)
         m_textures << texture;
-    } else {
-        std::swap(m_textures[plane], texture);
-        gl->glDeleteTextures(1, &texture);
-    }
+    else
+        m_textures[plane] = texture;
 }
 
 void QWaylandBufferMaterial::bind()
 {
-    QOpenGLFunctions *gl = QOpenGLContext::currentContext()->functions();
-    const GLenum target = bufferTypes[m_format].textureTarget;
-
     ensureTextures(bufferTypes[m_format].planeCount);
 
     switch (m_textures.size()) {
     case 3:
-        gl->glActiveTexture(GL_TEXTURE2);
-        gl->glBindTexture(target, m_textures[2]);
+        if (m_textures[2])
+            m_textures[2]->bind(GL_TEXTURE2);
     case 2:
-        gl->glActiveTexture(GL_TEXTURE1);
-        gl->glBindTexture(target, m_textures[1]);
+        if (m_textures[1])
+            m_textures[1]->bind(GL_TEXTURE1);
     case 1:
-        gl->glActiveTexture(GL_TEXTURE0);
-        gl->glBindTexture(target, m_textures[0]);
+        if (m_textures[0])
+            m_textures[0]->bind(GL_TEXTURE0);
     }
 }
 
@@ -257,15 +254,8 @@ void QWaylandBufferMaterial::setTextureParameters(GLenum target)
 //TODO move this into a separate centralized texture management class
 void QWaylandBufferMaterial::ensureTextures(int count)
 {
-    QOpenGLFunctions *gl = QOpenGLContext::currentContext()->functions();
-    const GLenum target = bufferTypes[m_format].textureTarget;
-    GLuint texture;
-
     for (int plane = m_textures.size(); plane < count; plane++) {
-        gl->glGenTextures(1, &texture);
-        gl->glBindTexture(target, texture);
-        setTextureParameters(target);
-        m_textures << texture;
+        m_textures << nullptr;
     }
 }
 
@@ -293,7 +283,7 @@ public:
         delete m_sgTex;
         m_sgTex = 0;
         if (m_ref.hasBuffer()) {
-            if (buffer.isShm()) {
+            if (buffer.isSharedMemory()) {
                 m_sgTex = surfaceItem->window()->createTextureFromImage(buffer.image());
                 if (m_sgTex) {
                     m_sgTex->bind();
@@ -305,17 +295,14 @@ public:
                     opt |= QQuickWindow::TextureHasAlphaChannel;
                 }
 
-                GLuint texture;
-                glGenTextures(1, &texture);
-                glBindTexture(GL_TEXTURE_2D, texture);
-                buffer.bindToTexture();
-                m_sgTex = surfaceItem->window()->createTextureFromId(texture , QSize(surfaceItem->width(), surfaceItem->height()), opt);
+                auto texture = buffer.toOpenGLTexture();
+                m_sgTex = surfaceItem->window()->createTextureFromId(texture->textureId() , QSize(surfaceItem->width(), surfaceItem->height()), opt);
             }
         }
         emit textureChanged();
     }
 
-    QSGTexture *texture() const Q_DECL_OVERRIDE
+    QSGTexture *texture() const override
     {
         if (m_sgTex)
             m_sgTex->setFiltering(m_smooth ? QSGTexture::Linear : QSGTexture::Nearest);
@@ -332,7 +319,7 @@ private:
 /*!
  * \qmltype WaylandQuickItem
  * \inqmlmodule QtWayland.Compositor
- * \preliminary
+ * \since 5.8
  * \brief Provides a Qt Quick item that represents a WaylandView.
  *
  * Qt Quick-based Wayland compositors can use this type to display a client's
@@ -343,7 +330,7 @@ private:
 /*!
  * \class QWaylandQuickItem
  * \inmodule QtWaylandCompositor
- * \preliminary
+ * \since 5.8
  * \brief The QWaylandQuickItem class provides a Qt Quick item representing a QWaylandView.
  *
  * When writing a QWaylandCompositor in Qt Quick, this class can be used to display a
@@ -492,13 +479,14 @@ void QWaylandQuickItem::mousePressEvent(QMouseEvent *event)
         return;
     }
 
-    QWaylandInputDevice *inputDevice = compositor()->inputDeviceFor(event);
+    QWaylandSeat *seat = compositor()->seatFor(event);
 
     if (d->focusOnClick)
-        takeFocus(inputDevice);
+        takeFocus(seat);
 
-    inputDevice->sendMouseMoveEvent(d->view.data(), event->localPos() / d->scaleFactor(), event->windowPos());
-    inputDevice->sendMousePressEvent(event->button());
+    seat->sendMouseMoveEvent(d->view.data(), mapToSurface(event->localPos()), event->windowPos());
+    seat->sendMousePressEvent(event->button());
+    d->hoverPos = event->pos();
 }
 
 /*!
@@ -508,8 +496,24 @@ void QWaylandQuickItem::mouseMoveEvent(QMouseEvent *event)
 {
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents()) {
-        QWaylandInputDevice *inputDevice = compositor()->inputDeviceFor(event);
-        inputDevice->sendMouseMoveEvent(d->view.data(), event->localPos() / d->scaleFactor(), event->windowPos());
+        QWaylandSeat *seat = compositor()->seatFor(event);
+#if QT_CONFIG(draganddrop)
+        if (d->isDragging) {
+            QWaylandQuickOutput *currentOutput = qobject_cast<QWaylandQuickOutput *>(view()->output());
+            //TODO: also check if dragging onto other outputs
+            QWaylandQuickItem *targetItem = qobject_cast<QWaylandQuickItem *>(currentOutput->pickClickableItem(mapToScene(event->localPos())));
+            QWaylandSurface *targetSurface = targetItem ? targetItem->surface() : nullptr;
+            if (targetSurface) {
+                QPointF position = mapToItem(targetItem, event->localPos());
+                QPointF surfacePosition = targetItem->mapToSurface(position);
+                seat->drag()->dragMove(targetSurface, surfacePosition);
+            }
+        } else
+#endif // QT_CONFIG(draganddrop)
+        {
+            seat->sendMouseMoveEvent(d->view.data(), mapToSurface(event->localPos()), event->windowPos());
+            d->hoverPos = event->pos();
+        }
     } else {
         emit mouseMove(event->windowPos());
         event->ignore();
@@ -523,8 +527,16 @@ void QWaylandQuickItem::mouseReleaseEvent(QMouseEvent *event)
 {
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents()) {
-        QWaylandInputDevice *inputDevice = compositor()->inputDeviceFor(event);
-        inputDevice->sendMouseReleaseEvent(event->button());
+        QWaylandSeat *seat = compositor()->seatFor(event);
+#if QT_CONFIG(draganddrop)
+        if (d->isDragging) {
+            d->isDragging = false;
+            seat->drag()->drop();
+        } else
+#endif
+        {
+            seat->sendMouseReleaseEvent(event->button());
+        }
     } else {
         emit mouseRelease();
         event->ignore();
@@ -542,8 +554,9 @@ void QWaylandQuickItem::hoverEnterEvent(QHoverEvent *event)
         return;
     }
     if (d->shouldSendInputEvents()) {
-        QWaylandInputDevice *inputDevice = compositor()->inputDeviceFor(event);
-        inputDevice->sendMouseMoveEvent(d->view.data(), event->pos(), mapToScene(event->pos()));
+        QWaylandSeat *seat = compositor()->seatFor(event);
+        seat->sendMouseMoveEvent(d->view.data(), event->pos(), mapToScene(event->pos()));
+        d->hoverPos = event->pos();
     } else {
         event->ignore();
     }
@@ -562,8 +575,11 @@ void QWaylandQuickItem::hoverMoveEvent(QHoverEvent *event)
         }
     }
     if (d->shouldSendInputEvents()) {
-        QWaylandInputDevice *inputDevice = compositor()->inputDeviceFor(event);
-        inputDevice->sendMouseMoveEvent(d->view.data(), event->pos() / d->scaleFactor(), mapToScene(event->pos()));
+        QWaylandSeat *seat = compositor()->seatFor(event);
+        if (event->pos() != d->hoverPos) {
+            seat->sendMouseMoveEvent(d->view.data(), mapToSurface(event->pos()), mapToScene(event->pos()));
+            d->hoverPos = event->pos();
+        }
     } else {
         event->ignore();
     }
@@ -576,13 +592,14 @@ void QWaylandQuickItem::hoverLeaveEvent(QHoverEvent *event)
 {
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents()) {
-        QWaylandInputDevice *inputDevice = compositor()->inputDeviceFor(event);
-        inputDevice->setMouseFocus(Q_NULLPTR);
+        QWaylandSeat *seat = compositor()->seatFor(event);
+        seat->setMouseFocus(Q_NULLPTR);
     } else {
         event->ignore();
     }
 }
 
+#if QT_CONFIG(wheelevent)
 /*!
  * \internal
  */
@@ -595,12 +612,13 @@ void QWaylandQuickItem::wheelEvent(QWheelEvent *event)
             return;
         }
 
-        QWaylandInputDevice *inputDevice = compositor()->inputDeviceFor(event);
-        inputDevice->sendMouseWheelEvent(event->orientation(), event->delta());
+        QWaylandSeat *seat = compositor()->seatFor(event);
+        seat->sendMouseWheelEvent(event->orientation(), event->delta());
     } else {
         event->ignore();
     }
 }
+#endif
 
 /*!
  * \internal
@@ -609,8 +627,8 @@ void QWaylandQuickItem::keyPressEvent(QKeyEvent *event)
 {
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents()) {
-        QWaylandInputDevice *inputDevice = compositor()->inputDeviceFor(event);
-        inputDevice->sendFullKeyEvent(event);
+        QWaylandSeat *seat = compositor()->seatFor(event);
+        seat->sendFullKeyEvent(event);
     } else {
         event->ignore();
     }
@@ -623,8 +641,8 @@ void QWaylandQuickItem::keyReleaseEvent(QKeyEvent *event)
 {
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents() && hasFocus()) {
-        QWaylandInputDevice *inputDevice = compositor()->inputDeviceFor(event);
-        inputDevice->sendFullKeyEvent(event);
+        QWaylandSeat *seat = compositor()->seatFor(event);
+        seat->sendFullKeyEvent(event);
     } else {
         event->ignore();
     }
@@ -637,19 +655,7 @@ void QWaylandQuickItem::touchEvent(QTouchEvent *event)
 {
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents() && d->touchEventsEnabled) {
-        QWaylandInputDevice *inputDevice = compositor()->inputDeviceFor(event);
-
-        if (event->type() == QEvent::TouchBegin) {
-            QQuickItem *grabber = window()->mouseGrabberItem();
-            if (grabber != this)
-                grabMouse();
-        }
-
-        if (event->type() == QEvent::TouchEnd) {
-            QQuickItem *grabber = window()->mouseGrabberItem();
-            if (grabber == this)
-                ungrabMouse();
-        }
+        QWaylandSeat *seat = compositor()->seatFor(event);
 
         QPoint pointPos;
         const QList<QTouchEvent::TouchPoint> &points = event->touchPoints();
@@ -662,27 +668,36 @@ void QWaylandQuickItem::touchEvent(QTouchEvent *event)
         }
 
         event->accept();
-        if (inputDevice->mouseFocus() != d->view.data()) {
-            inputDevice->sendMouseMoveEvent(d->view.data(), pointPos, mapToScene(pointPos));
+        if (seat->mouseFocus() != d->view.data()) {
+            seat->sendMouseMoveEvent(d->view.data(), pointPos, mapToScene(pointPos));
         }
-        inputDevice->sendFullTouchEvent(event);
+        seat->sendFullTouchEvent(surface(), event);
+
+        if (event->type() == QEvent::TouchBegin) {
+            d->touchingSeats.append(seat);
+        } else if (event->type() == QEvent::TouchEnd || event->type() == QEvent::TouchCancel) {
+            d->touchingSeats.removeOne(seat);
+        }
+
+        if (event->type() == QEvent::TouchBegin && d->focusOnClick)
+            takeFocus(seat);
     } else {
         event->ignore();
     }
 }
 
-/*!
- * \internal
- */
-void QWaylandQuickItem::mouseUngrabEvent()
+void QWaylandQuickItem::touchUngrabEvent()
 {
-    if (surface()) {
-        QTouchEvent e(QEvent::TouchCancel);
-        touchEvent(&e);
-    }
+    Q_D(QWaylandQuickItem);
+
+    if (d->shouldSendInputEvents())
+        for (auto seat : d->touchingSeats)
+            seat->sendTouchCancelEvent(surface()->client());
+
+    d->touchingSeats.clear();
 }
 
-#ifndef QT_NO_IM
+#if QT_CONFIG(im)
 /*!
  * \internal
  */
@@ -762,6 +777,80 @@ void QWaylandQuickItem::setSubsurfaceHandler(QObject *handler)
     }
 }
 
+/*!
+ * \property QWaylandQuickItem::output
+ *
+ * This property holds the output on which this item is displayed.
+ */
+QWaylandOutput *QWaylandQuickItem::output() const
+{
+    Q_D(const QWaylandQuickItem);
+    return d->view->output();
+}
+
+void QWaylandQuickItem::setOutput(QWaylandOutput *output)
+{
+    Q_D(QWaylandQuickItem);
+    d->view->setOutput(output);
+}
+
+/*!
+ * \property QWaylandQuickItem::bufferLocked
+ *
+ * This property holds whether the item's buffer is currently locked. As long as
+ * the buffer is locked, it will not be released and returned to the client.
+ *
+ * The default is false.
+ */
+bool QWaylandQuickItem::isBufferLocked() const
+{
+    Q_D(const QWaylandQuickItem);
+    return d->view->isBufferLocked();
+}
+
+void QWaylandQuickItem::setBufferLocked(bool locked)
+{
+    Q_D(QWaylandQuickItem);
+    d->view->setBufferLocked(locked);
+}
+
+/*!
+ * \property bool QWaylandQuickItem::allowDiscardFrontBuffer
+ *
+ * By default, the item locks the current buffer until a new buffer is available
+ * and updatePaintNode() is called. Set this property to true to allow Qt to release the buffer
+ * immediately when the throttling view is no longer using it. This is useful for items that have
+ * slow update intervals.
+ */
+bool QWaylandQuickItem::allowDiscardFrontBuffer() const
+{
+    Q_D(const QWaylandQuickItem);
+    return d->view->allowDiscardFrontBuffer();
+}
+
+void QWaylandQuickItem::setAllowDiscardFrontBuffer(bool discard)
+{
+    Q_D(QWaylandQuickItem);
+    d->view->setAllowDiscardFrontBuffer(discard);
+}
+
+/*!
+ * \qmlmethod QtWaylandCompositor::WaylandQuickItem::setPrimary
+ *
+ * Makes this WaylandQuickItem the primary view for the surface.
+ */
+
+/*!
+ * Makes this QWaylandQuickItem's view the primary view for the surface.
+ *
+ * \sa QWaylandSurface::primaryView
+ */
+void QWaylandQuickItem::setPrimary()
+{
+    Q_D(QWaylandQuickItem);
+    d->view->setPrimary();
+}
+
 /*!
  * \internal
  */
@@ -769,26 +858,32 @@ void QWaylandQuickItem::handleSurfaceChanged()
 {
     Q_D(QWaylandQuickItem);
     if (d->oldSurface) {
-        disconnect(d->oldSurface, &QWaylandSurface::mappedChanged, this, &QWaylandQuickItem::surfaceMappedChanged);
+        disconnect(d->oldSurface, &QWaylandSurface::hasContentChanged, this, &QWaylandQuickItem::surfaceMappedChanged);
         disconnect(d->oldSurface, &QWaylandSurface::parentChanged, this, &QWaylandQuickItem::parentChanged);
         disconnect(d->oldSurface, &QWaylandSurface::sizeChanged, this, &QWaylandQuickItem::updateSize);
         disconnect(d->oldSurface, &QWaylandSurface::bufferScaleChanged, this, &QWaylandQuickItem::updateSize);
         disconnect(d->oldSurface, &QWaylandSurface::configure, this, &QWaylandQuickItem::updateBuffer);
         disconnect(d->oldSurface, &QWaylandSurface::redraw, this, &QQuickItem::update);
         disconnect(d->oldSurface, &QWaylandSurface::childAdded, this, &QWaylandQuickItem::handleSubsurfaceAdded);
-#ifndef QT_NO_IM
+#if QT_CONFIG(draganddrop)
+        disconnect(d->oldSurface, &QWaylandSurface::dragStarted, this, &QWaylandQuickItem::handleDragStarted);
+#endif
+#if QT_CONFIG(im)
         disconnect(d->oldSurface->inputMethodControl(), &QWaylandInputMethodControl::updateInputMethod, this, &QWaylandQuickItem::updateInputMethod);
 #endif
     }
     if (QWaylandSurface *newSurface = d->view->surface()) {
-        connect(newSurface, &QWaylandSurface::mappedChanged, this, &QWaylandQuickItem::surfaceMappedChanged);
+        connect(newSurface, &QWaylandSurface::hasContentChanged, this, &QWaylandQuickItem::surfaceMappedChanged);
         connect(newSurface, &QWaylandSurface::parentChanged, this, &QWaylandQuickItem::parentChanged);
         connect(newSurface, &QWaylandSurface::sizeChanged, this, &QWaylandQuickItem::updateSize);
         connect(newSurface, &QWaylandSurface::bufferScaleChanged, this, &QWaylandQuickItem::updateSize);
         connect(newSurface, &QWaylandSurface::configure, this, &QWaylandQuickItem::updateBuffer);
         connect(newSurface, &QWaylandSurface::redraw, this, &QQuickItem::update);
         connect(newSurface, &QWaylandSurface::childAdded, this, &QWaylandQuickItem::handleSubsurfaceAdded);
-#ifndef QT_NO_IM
+#if QT_CONFIG(draganddrop)
+        connect(newSurface, &QWaylandSurface::dragStarted, this, &QWaylandQuickItem::handleDragStarted);
+#endif
+#if QT_CONFIG(im)
         connect(newSurface->inputMethodControl(), &QWaylandInputMethodControl::updateInputMethod, this, &QWaylandQuickItem::updateInputMethod);
 #endif
 
@@ -805,7 +900,7 @@ void QWaylandQuickItem::handleSurfaceChanged()
     }
     surfaceChangedEvent(d->view->surface(), d->oldSurface);
     d->oldSurface = d->view->surface();
-#ifndef QT_NO_IM
+#if QT_CONFIG(im)
     updateInputMethod(Qt::ImQueryInput);
 #endif
 }
@@ -814,16 +909,16 @@ void QWaylandQuickItem::handleSurfaceChanged()
  * Calling this function causes the item to take the focus of the
  * input \a device.
  */
-void QWaylandQuickItem::takeFocus(QWaylandInputDevice *device)
+void QWaylandQuickItem::takeFocus(QWaylandSeat *device)
 {
     forceActiveFocus();
 
     if (!surface())
         return;
 
-    QWaylandInputDevice *target = device;
+    QWaylandSeat *target = device;
     if (!target) {
-        target = compositor()->defaultInputDevice();
+        target = compositor()->defaultSeat();
     }
     target->setKeyboardFocus(surface());
     QWaylandTextInput *textInput = QWaylandTextInput::findIn(target);
@@ -869,7 +964,7 @@ void QWaylandQuickItem::updateSize()
  * \qmlproperty bool QtWaylandCompositor::WaylandQuickItem::focusOnClick
  *
  * This property specifies whether the WaylandQuickItem should take focus when
- * it is clicked.
+ * it is clicked or touched.
  *
  * The default is \c true.
  */
@@ -878,7 +973,7 @@ void QWaylandQuickItem::updateSize()
  * \property QWaylandQuickItem::focusOnClick
  *
  * This property specifies whether the QWaylandQuickItem should take focus when
- * it is clicked.
+ * it is clicked or touched.
  *
  * The default is \c true.
  */
@@ -904,12 +999,22 @@ void QWaylandQuickItem::setFocusOnClick(bool focus)
  */
 bool QWaylandQuickItem::inputRegionContains(const QPointF &localPosition)
 {
-    Q_D(QWaylandQuickItem);
     if (QWaylandSurface *s = surface())
-        return s->inputRegionContains(localPosition.toPoint() / d->scaleFactor());
+        return s->inputRegionContains(mapToSurface(localPosition).toPoint());
     return false;
 }
 
+/*!
+ * Maps the given \a point in this item's coordinate system to the equivalent
+ * point within the Wayland surface's coordinate system, and returns the mapped
+ * coordinate.
+ */
+QPointF QWaylandQuickItem::mapToSurface(const QPointF &point) const
+{
+    Q_D(const QWaylandQuickItem);
+    return point / d->scaleFactor();
+}
+
 /*!
  * \qmlproperty bool QtWaylandCompositor::WaylandQuickItem::sizeFollowsSurface
  *
@@ -942,7 +1047,7 @@ void QWaylandQuickItem::setSizeFollowsSurface(bool sizeFollowsSurface)
     emit sizeFollowsSurfaceChanged();
 }
 
-#ifndef QT_NO_IM
+#if QT_CONFIG(im)
 QVariant QWaylandQuickItem::inputMethodQuery(Qt::InputMethodQuery query) const
 {
     return inputMethodQuery(query, QVariant());
@@ -965,14 +1070,14 @@ QVariant QWaylandQuickItem::inputMethodQuery(Qt::InputMethodQuery query, QVarian
 /*!
     \qmlproperty bool QtWaylandCompositor::WaylandQuickItem::paintEnabled
 
-    If this property is \c true, the item is hidden, though the texture
+    Returns true if the item is hidden, though the texture
     is still updated. As opposed to hiding the item by
     setting \l{Item::visible}{visible} to \c false, setting this property to \c true
     will not prevent mouse or keyboard input from reaching item.
 */
 
 /*!
-    If this property is \c true, the item is hidden, though the texture
+    Returns true if the item is hidden, though the texture
     is still updated. As opposed to hiding the item by
     setting \l{Item::visible}{visible} to \c false, setting this property to \c true
     will not prevent mouse or keyboard input from reaching item.
@@ -1009,11 +1114,16 @@ void QWaylandQuickItem::updateBuffer(bool hasBuffer)
 void QWaylandQuickItem::updateWindow()
 {
     Q_D(QWaylandQuickItem);
+
+    QQuickWindow *newWindow = window();
+    if (newWindow == d->connectedWindow)
+        return;
+
     if (d->connectedWindow) {
         disconnect(d->connectedWindow, &QQuickWindow::beforeSynchronizing, this, &QWaylandQuickItem::beforeSync);
     }
 
-    d->connectedWindow = window();
+    d->connectedWindow = newWindow;
 
     if (d->connectedWindow) {
         connect(d->connectedWindow, &QQuickWindow::beforeSynchronizing, this, &QWaylandQuickItem::beforeSync, Qt::DirectConnection);
@@ -1035,7 +1145,7 @@ void QWaylandQuickItem::beforeSync()
     }
 }
 
-#ifndef QT_NO_IM
+#if QT_CONFIG(im)
 void QWaylandQuickItem::updateInputMethod(Qt::InputMethodQueries queries)
 {
     Q_D(QWaylandQuickItem);
@@ -1049,9 +1159,12 @@ void QWaylandQuickItem::updateInputMethod(Qt::InputMethodQueries queries)
 QSGNode *QWaylandQuickItem::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeData *)
 {
     Q_D(QWaylandQuickItem);
-    const bool mapped = surface() && surface()->isMapped() && d->view->currentBuffer().hasBuffer();
+    const bool bufferHasContent = d->view->currentBuffer().hasContent();
 
-    if (!mapped || !d->paintEnabled) {
+    if (d->view->isBufferLocked() && !bufferHasContent && d->paintEnabled)
+        return oldNode;
+
+    if (!bufferHasContent || !d->paintEnabled) {
         delete oldNode;
         return 0;
     }
@@ -1061,7 +1174,8 @@ QSGNode *QWaylandQuickItem::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeDat
     const QRectF rect = invertY ? QRectF(0, height(), width(), -height())
                                 : QRectF(0, 0, width(), height());
 
-    if (ref.isShm() || bufferTypes[ref.bufferFormatEgl()].canProvideTexture) {
+    if (ref.isSharedMemory() || bufferTypes[ref.bufferFormatEgl()].canProvideTexture) {
+        // This case could covered by the more general path below, but this is more efficient (especially when using ShaderEffect items).
         QSGSimpleTextureNode *node = static_cast<QSGSimpleTextureNode *>(oldNode);
 
         if (!node) {
@@ -1104,13 +1218,11 @@ QSGNode *QWaylandQuickItem::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeDat
         if (d->newTexture) {
             d->newTexture = false;
             for (int plane = 0; plane < bufferTypes[ref.bufferFormatEgl()].planeCount; plane++)
-                if (uint texture = ref.textureForPlane(plane))
+                if (auto texture = ref.toOpenGLTexture(plane))
                     material->setTextureForPlane(plane, texture);
             material->bind();
-            ref.bindToTexture();
         }
 
-        ref.updateTexture();
         QSGGeometry::updateTexturedRectGeometry(geometry, rect, QRectF(0, 0, 1, 1));
 
         node->setGeometry(geometry);
@@ -1144,6 +1256,8 @@ void QWaylandQuickItem::setInputEventsEnabled(bool enabled)
 {
     Q_D(QWaylandQuickItem);
     if (d->inputEventsEnabled != enabled) {
+        if (enabled)
+            setEnabled(true);
         d->setInputEventsEnabled(enabled);
         emit inputEventsEnabledChanged();
     }
@@ -1178,5 +1292,20 @@ void QWaylandQuickItem::handleSubsurfacePosition(const QPoint &pos)
     QQuickItem::setPosition(pos * d->scaleFactor());
 }
 
-QT_END_NAMESPACE
+#if QT_CONFIG(draganddrop)
+void QWaylandQuickItem::handleDragStarted(QWaylandDrag *drag)
+{
+    Q_D(QWaylandQuickItem);
+    Q_ASSERT(drag->origin() == surface());
+    drag->seat()->setMouseFocus(nullptr);
+    d->isDragging = true;
+}
+#endif
 
+qreal QWaylandQuickItemPrivate::scaleFactor() const
+{
+    return (view->output() ? view->output()->scaleFactor() : 1)
+            / (window ? window->devicePixelRatio() : 1);
+}
+
+QT_END_NAMESPACE
